@@ -1,6 +1,7 @@
 /*
 Copyright (C) 1996-2001 Id Software, Inc.
 Copyright (C) 2002-2009 John Fitzgibbons and others
+Copyright (C) 2010-2014 QuakeSpasm developers
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -45,9 +46,6 @@ vec3_t	vpn;
 vec3_t	vright;
 vec3_t	r_origin;
 
-float	r_world_matrix[16];
-float	r_base_world_matrix[16];
-
 float r_fovx, r_fovy; //johnfitz -- rendering fov may be different becuase of r_waterwarp and r_stereo
 
 //
@@ -98,13 +96,142 @@ cvar_t	r_showbboxes = {"r_showbboxes", "0", CVAR_NONE};
 cvar_t	r_lerpmodels = {"r_lerpmodels", "1", CVAR_NONE};
 cvar_t	r_lerpmove = {"r_lerpmove", "1", CVAR_NONE};
 cvar_t	r_nolerp_list = {"r_nolerp_list", "progs/flame.mdl,progs/flame2.mdl,progs/braztall.mdl,progs/brazshrt.mdl,progs/longtrch.mdl,progs/flame_pyre.mdl,progs/v_saw.mdl,progs/v_xfist.mdl,progs/h2stuff/newfire.mdl", CVAR_NONE};
+cvar_t	r_noshadow_list = {"r_noshadow_list", "progs/flame2.mdl,progs/flame.mdl,progs/bolt1.mdl,progs/bolt2.mdl,progs/bolt3.mdl,progs/laser.mdl", CVAR_NONE};
+
 extern cvar_t	r_vfog;
 //johnfitz
 
-cvar_t	gl_zfix = {"gl_zfix", "1", CVAR_ARCHIVE}; // QuakeSpasm z-fighting fix
+cvar_t	gl_zfix = {"gl_zfix", "0", CVAR_NONE}; // QuakeSpasm z-fighting fix
 
 qboolean r_drawflat_cheatsafe, r_fullbright_cheatsafe, r_lightmap_cheatsafe, r_drawworld_cheatsafe; //johnfitz
 
+//==============================================================================
+//
+// GLSL GAMMA CORRECTION
+//
+//==============================================================================
+
+static GLuint r_gamma_texture;
+static GLuint r_gamma_program;
+
+// uniforms used in gamma shader
+static GLuint gammaLoc;
+static GLuint textureLoc;
+
+/*
+=============
+GLSLGamma_DeleteTexture
+=============
+*/
+void GLSLGamma_DeleteTexture (void)
+{
+	glDeleteTextures (1, &r_gamma_texture);
+	r_gamma_texture = 0;
+	r_gamma_program = 0; // deleted in R_DeleteShaders
+}
+
+/*
+=============
+GLSLGamma_CreateShaders
+=============
+*/
+static void GLSLGamma_CreateShaders (void)
+{
+	const GLchar *vertSource = \
+		"#version 110\n"
+		"\n"
+		"void main(void) {\n"
+		"	gl_Position = vec4(gl_Vertex.xy, 0.0, 1.0);\n"
+		"	gl_TexCoord[0] = gl_MultiTexCoord0;\n"
+		"}\n";
+
+	const GLchar *fragSource = \
+		"#version 110\n"
+		"\n"
+		"uniform sampler2D GammaTexture;\n"
+		"uniform float GammaValue;\n"
+		"\n"
+		"void main(void) {\n"
+		"	  vec4 frag = texture2D(GammaTexture, gl_TexCoord[0].xy);\n"
+		"	  gl_FragColor = vec4(pow(frag.rgb, vec3(GammaValue)), 1.0);\n"
+		"}\n";
+
+	if (!gl_glsl_gamma_able)
+		return;
+
+	r_gamma_program = GL_CreateProgram (vertSource, fragSource, 0, NULL);
+
+// get uniform locations
+	gammaLoc = GL_GetUniformLocation (&r_gamma_program, "GammaValue");
+	textureLoc = GL_GetUniformLocation (&r_gamma_program, "GammaTexture");
+}
+
+/*
+=============
+GLSLGamma_GammaCorrect
+=============
+*/
+void GLSLGamma_GammaCorrect (void)
+{
+	if (!gl_glsl_gamma_able)
+		return;
+
+	if (vid_gamma.value == 1)
+		return;
+
+// create render-to-texture texture if needed
+	if (!r_gamma_texture)
+	{
+		glGenTextures (1, &r_gamma_texture);
+		glBindTexture (GL_TEXTURE_2D, r_gamma_texture);
+	
+		glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA8, glwidth, glheight, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, NULL);
+		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
+		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+	}
+
+// create shader if needed
+	if (!r_gamma_program)
+	{
+		GLSLGamma_CreateShaders ();
+		if (!r_gamma_program)
+		{
+			Sys_Error("GLSLGamma_CreateShaders failed");
+		}
+	}
+	
+// copy the framebuffer to the texture
+	GL_DisableMultitexture();
+	glBindTexture (GL_TEXTURE_2D, r_gamma_texture);
+	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, glx, gly, glwidth, glheight);
+
+// draw the texture back to the framebuffer with a fragment shader
+	GL_UseProgramFunc (r_gamma_program);
+	GL_Uniform1fFunc (gammaLoc, vid_gamma.value);
+	GL_Uniform1iFunc (textureLoc, 0); // use texture unit 0
+
+	glDisable (GL_ALPHA_TEST);
+	glDisable (GL_DEPTH_TEST);
+
+	glViewport (glx, gly, glwidth, glheight);
+
+	glBegin (GL_QUADS);
+	glTexCoord2f (0, 0);
+	glVertex2f (-1, -1);
+	glTexCoord2f (1, 0);
+	glVertex2f (1, -1);
+	glTexCoord2f (1, 1);
+	glVertex2f (1, 1);
+	glTexCoord2f (0, 1);
+	glVertex2f (-1, 1);
+	glEnd ();
+	
+	GL_UseProgramFunc (0);
+	
+// clear cached binding
+	GL_Bind (notexture);
+}
+
 /*
 =================
 R_CullBox -- johnfitz -- replaced with new function from lordhavoc
@@ -341,8 +468,6 @@ void R_SetupGL (void)
     glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
     glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);
 
-	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);
-
 	//
 	// set drawing parms
 	//
@@ -366,6 +491,9 @@ void R_Clear (void)
 	unsigned int clearbits;
 
 	clearbits = GL_DEPTH_BUFFER_BIT;
+	// from mh -- if we get a stencil buffer, we should clear it, even though we don't use it
+	if (gl_stencilbits)
+		clearbits |= GL_STENCIL_BUFFER_BIT;
 	if (gl_clear.value)
 		clearbits |= GL_COLOR_BUFFER_BIT;
 	glClear (clearbits);
@@ -451,6 +579,11 @@ void R_SetupView (void)
 //
 //==============================================================================
 
+static int R_EntitySortFunc(const void *a, const void *b)
+{
+	return (*(entity_t **)a)->model - (*(entity_t **)b)->model;
+}
+
 /*
 =============
 R_DrawEntitiesOnList
@@ -458,15 +591,21 @@ R_DrawEntitiesOnList
 */
 void R_DrawEntitiesOnList (qboolean alphapass) //johnfitz -- added parameter
 {
+	entity_t	*entities_sorted[MAX_VISEDICTS];
 	int		i;
 
 	if (!r_drawentities.value)
 		return;
 
+	//ericw -- draw the entities sorted by model, to eliminate redundant texture binding
+	memcpy (entities_sorted, cl_visedicts, cl_numvisedicts * sizeof(entity_t *));
+	qsort (entities_sorted, cl_numvisedicts, sizeof(entity_t *), R_EntitySortFunc);
+	//ericw --
+
 	//johnfitz -- sprites are not a special case
 	for (i=0 ; i<cl_numvisedicts ; i++)
 	{
-		currententity = cl_visedicts[i];
+		currententity = entities_sorted[i];
 
 		//johnfitz -- if alphapass is true, draw only alpha entites this time
 		//if alphapass is false, draw only nonalpha entities this time
@@ -643,7 +782,7 @@ void R_ShowTris (void)
 
 	if (r_drawworld.value)
 	{
-		R_DrawTextureChains_ShowTris ();
+		R_DrawWorld_ShowTris ();
 	}
 
 	if (r_drawentities.value)
@@ -715,6 +854,15 @@ void R_DrawShadows (void)
 	if (!r_shadows.value || !r_drawentities.value || r_drawflat_cheatsafe || r_lightmap_cheatsafe)
 		return;
 
+	// Use stencil buffer to prevent self-intersecting shadows, from Baker (MarkV)
+	if (gl_stencilbits)
+	{
+		glClear(GL_STENCIL_BUFFER_BIT);
+		glStencilFunc(GL_EQUAL, 0, ~0);
+		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
+		glEnable(GL_STENCIL_TEST);
+	}
+
 	for (i=0 ; i<cl_numvisedicts ; i++)
 	{
 		currententity = cl_visedicts[i];
@@ -727,6 +875,11 @@ void R_DrawShadows (void)
 
 		GL_DrawAliasShadow (currententity);
 	}
+
+	if (gl_stencilbits)
+	{
+		glDisable(GL_STENCIL_TEST);
+	}
 }
 
 /*
@@ -750,7 +903,7 @@ void R_RenderScene (void)
 
 	R_DrawEntitiesOnList (false); //johnfitz -- false means this is the pass for nonalpha entities
 
-	R_DrawTextureChains_Water (); //johnfitz -- drawn here since they might have transparency
+	R_DrawWorld_Water (); //johnfitz -- drawn here since they might have transparency
 
 	R_DrawEntitiesOnList (true); //johnfitz -- true means this is the pass for alpha entities
 
