@@ -2,6 +2,7 @@
 Copyright (C) 1996-2001 Id Software, Inc.
 Copyright (C) 2002-2009 John Fitzgibbons and others
 Copyright (C) 2007-2008 Kristian Duske
+Copyright (C) 2010-2014 QuakeSpasm developers
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -68,7 +69,6 @@ console is:
 	half
 	full
 
-
 */
 
 
@@ -183,6 +183,8 @@ void SCR_DrawCenterString (void) //actually do the drawing
 		y = 200*0.35;	//johnfitz -- 320x200 coordinate system
 	else
 		y = 48;
+	if (crosshair.value)
+		y -= 8;
 
 	do
 	{
@@ -311,17 +313,17 @@ static void SCR_CalcRefdef (void)
 	else
 		sb_lines = 48 * scale;
 
-	size = fmin(scr_viewsize.value, 100) / 100;
+	size = q_min(scr_viewsize.value, 100) / 100;
 	//johnfitz
 
 	//johnfitz -- rewrote this section
-	r_refdef.vrect.width = fmax(glwidth * size, 96); //no smaller than 96, for icons
-	r_refdef.vrect.height = fmin(glheight * size, glheight - sb_lines); //make room for sbar
+	r_refdef.vrect.width = q_max(glwidth * size, 96); //no smaller than 96, for icons
+	r_refdef.vrect.height = q_min(glheight * size, glheight - sb_lines); //make room for sbar
 	r_refdef.vrect.x = (glwidth - r_refdef.vrect.width)/2;
 	r_refdef.vrect.y = (glheight - sb_lines - r_refdef.vrect.height)/2;
 	//johnfitz
 
-	r_refdef.fov_x = AdaptFovx(scr_fov.value, r_refdef.vrect.width, r_refdef.vrect.height);
+	r_refdef.fov_x = AdaptFovx(scr_fov.value, vid.width, vid.height);
 	r_refdef.fov_y = CalcFovy (r_refdef.fov_x, r_refdef.vrect.width, r_refdef.vrect.height);
 
 	scr_vrect = r_refdef.vrect;
@@ -385,6 +387,7 @@ void SCR_LoadPics (void)
 	scr_net = Draw_PicFromWad ("net");
 	scr_turtle = Draw_PicFromWad ("turtle");
 }
+
 /*
 ==================
 SCR_Init
@@ -472,7 +475,6 @@ void SCR_DrawFPS (void)
 		Draw_String (x, y, st);
 		scr_tileclear_updates = 0;
 	}
-
 }
 
 /*
@@ -707,14 +709,15 @@ void SCR_SetUpToDrawConsole (void)
 
 	if (scr_conlines < scr_con_current)
 	{
-		scr_con_current -= scr_conspeed.value*host_frametime/timescale; //johnfitz -- timescale
+		// ericw -- (glheight/600.0) factor makes conspeed resolution independent, using 800x600 as a baseline
+		scr_con_current -= scr_conspeed.value*(glheight/600.0)*host_frametime/timescale; //johnfitz -- timescale
 		if (scr_conlines > scr_con_current)
 			scr_con_current = scr_conlines;
-
 	}
 	else if (scr_conlines > scr_con_current)
 	{
-		scr_con_current += scr_conspeed.value*host_frametime/timescale; //johnfitz -- timescale
+		// ericw -- (glheight/600.0)
+		scr_con_current += scr_conspeed.value*(glheight/600.0)*host_frametime/timescale; //johnfitz -- timescale
 		if (scr_conlines < scr_con_current)
 			scr_con_current = scr_conlines;
 	}
@@ -749,7 +752,7 @@ void SCR_DrawConsole (void)
 /*
 ==============================================================================
 
-						SCREEN SHOTS
+SCREEN SHOTS
 
 ==============================================================================
 */
@@ -778,10 +781,16 @@ void SCR_ScreenShot_f (void)
 	{
 		Con_Printf ("SCR_ScreenShot_f: Couldn't find an unused filename\n");
 		return;
- 	}
+	}
 
 //get data
-	buffer = (byte *) malloc(glwidth*glheight*3);
+	if (!(buffer = (byte *) malloc(glwidth*glheight*3)))
+	{
+		Con_Printf ("SCR_ScreenShot_f: Couldn't allocate memory\n");
+		return;
+	}
+
+	glPixelStorei (GL_PACK_ALIGNMENT, 1);/* for widths that aren't a multiple of 4 */
 	glReadPixels (glx, gly, glwidth, glheight, GL_RGB, GL_UNSIGNED_BYTE, buffer);
 
 // now write the file
@@ -888,6 +897,7 @@ keypress.
 int SCR_ModalMessage (const char *text, float timeout) //johnfitz -- timeout
 {
 	double time1, time2; //johnfitz -- timeout
+	int lastkey, lastchar;
 
 	if (cls.state == ca_dedicated)
 		return true;
@@ -904,20 +914,18 @@ int SCR_ModalMessage (const char *text, float timeout) //johnfitz -- timeout
 	time1 = Sys_DoubleTime () + timeout; //johnfitz -- timeout
 	time2 = 0.0f; //johnfitz -- timeout
 
+	Key_BeginInputGrab ();
 	do
 	{
-		key_count = -1;		// wait for a key down and up
 		Sys_SendKeyEvents ();
-		Sys_Sleep(16);
+		Key_GetGrabbedInput (&lastkey, &lastchar);
+		Sys_Sleep (16);
 		if (timeout) time2 = Sys_DoubleTime (); //johnfitz -- zero timeout means wait forever.
-	} while (key_lastpress != 'y' &&
-			 key_lastpress != 'n' &&
-			 key_lastpress != K_ESCAPE &&
-			 time2 <= time1);
-
-	// make sure we don't ignore the next keypress
-	if (key_count < 0)
-		key_count = 0;
+	} while (lastchar != 'y' && lastchar != 'Y' &&
+		 lastchar != 'n' && lastchar != 'N' &&
+		 lastkey != K_ESCAPE &&
+		 time2 <= time1);
+	Key_EndInputGrab ();
 
 //	SCR_UpdateScreen (); //johnfitz -- commented out
 
@@ -926,7 +934,7 @@ int SCR_ModalMessage (const char *text, float timeout) //johnfitz -- timeout
 		return false;
 	//johnfitz
 
-	return key_lastpress == 'y';
+	return (lastchar == 'y' || lastchar == 'Y');
 }
 
 
@@ -937,9 +945,10 @@ int SCR_ModalMessage (const char *text, float timeout) //johnfitz -- timeout
 
 /*
 ==================
-SCR_TileClear -- johnfitz -- modified to use glwidth/glheight instead of vid.width/vid.height
-                             also fixed the dimentions of right and top panels
-							 also added scr_tileclear_updates
+SCR_TileClear
+johnfitz -- modified to use glwidth/glheight instead of vid.width/vid.height
+	    also fixed the dimentions of right and top panels
+	    also added scr_tileclear_updates
 ==================
 */
 void SCR_TileClear (void)
@@ -1068,6 +1077,8 @@ void SCR_UpdateScreen (void)
 
 	V_UpdateBlend (); //johnfitz -- V_UpdatePalette cleaned up and renamed
 
+	GLSLGamma_GammaCorrect ();
+
 	GL_EndRendering ();
 }
 
