@@ -2,7 +2,7 @@
 Copyright (C) 1996-2001 Id Software, Inc.
 Copyright (C) 2002-2009 John Fitzgibbons and others
 Copyright (C) 2007-2008 Kristian Duske
-Copyright (C) 2010 Ozkan Sezer and Steven Atkinson
+Copyright (C) 2010-2014 QuakeSpasm developers
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -27,11 +27,20 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "bgmusic.h"
 #include "resource.h"
 #if defined(SDL_FRAMEWORK) || defined(NO_SDL_CONFIG)
+#if defined(USE_SDL2)
+#include <SDL2/SDL.h>
+#else
 #include <SDL/SDL.h>
+#endif
 #else
 #include "SDL.h"
 #endif
 
+//ericw -- for putting the driver into multithreaded mode
+#ifdef __APPLE__
+#include <OpenGL/OpenGL.h>
+#endif
+
 #define MAX_MODE_LIST	600 //johnfitz -- was 30
 #define MAX_BPPS_LIST	5
 #define WARP_WIDTH		320
@@ -50,6 +59,8 @@ typedef struct {
 static const char *gl_vendor;
 static const char *gl_renderer;
 static const char *gl_version;
+static int gl_version_major;
+static int gl_version_minor;
 static const char *gl_extensions;
 static char * gl_extensions_nice;
 
@@ -58,7 +69,12 @@ static int		nummodes;
 
 static qboolean	vid_initialized = false;
 
+#if defined(USE_SDL2)
+static SDL_Window	*draw_context;
+static SDL_GLContext	gl_context;
+#else
 static SDL_Surface	*draw_context;
+#endif
 
 static qboolean	vid_locked = false; //johnfitz
 static qboolean	vid_changed = false;
@@ -82,9 +98,46 @@ qboolean gl_texture_env_add = false; //johnfitz
 qboolean gl_swap_control = false; //johnfitz
 qboolean gl_anisotropy_able = false; //johnfitz
 float gl_max_anisotropy; //johnfitz
+qboolean gl_texture_NPOT = false; //ericw
+qboolean gl_vbo_able = false; //ericw
+qboolean gl_glsl_able = false; //ericw
+GLint gl_max_texture_units = 0; //ericw
+qboolean gl_glsl_gamma_able = false; //ericw
+qboolean gl_glsl_alias_able = false; //ericw
+int gl_stencilbits;
 
 PFNGLMULTITEXCOORD2FARBPROC GL_MTexCoord2fFunc = NULL; //johnfitz
 PFNGLACTIVETEXTUREARBPROC GL_SelectTextureFunc = NULL; //johnfitz
+PFNGLCLIENTACTIVETEXTUREARBPROC GL_ClientActiveTextureFunc = NULL; //ericw
+PFNGLBINDBUFFERARBPROC GL_BindBufferFunc = NULL; //ericw
+PFNGLBUFFERDATAARBPROC GL_BufferDataFunc = NULL; //ericw
+PFNGLBUFFERSUBDATAARBPROC GL_BufferSubDataFunc = NULL; //ericw
+PFNGLDELETEBUFFERSARBPROC GL_DeleteBuffersFunc = NULL; //ericw
+PFNGLGENBUFFERSARBPROC GL_GenBuffersFunc = NULL; //ericw
+
+QS_PFNGLCREATESHADERPROC GL_CreateShaderFunc = NULL; //ericw
+QS_PFNGLDELETESHADERPROC GL_DeleteShaderFunc = NULL; //ericw
+QS_PFNGLDELETEPROGRAMPROC GL_DeleteProgramFunc = NULL; //ericw
+QS_PFNGLSHADERSOURCEPROC GL_ShaderSourceFunc = NULL; //ericw
+QS_PFNGLCOMPILESHADERPROC GL_CompileShaderFunc = NULL; //ericw
+QS_PFNGLGETSHADERIVPROC GL_GetShaderivFunc = NULL; //ericw
+QS_PFNGLGETSHADERINFOLOGPROC GL_GetShaderInfoLogFunc = NULL; //ericw
+QS_PFNGLGETPROGRAMIVPROC GL_GetProgramivFunc = NULL; //ericw
+QS_PFNGLGETPROGRAMINFOLOGPROC GL_GetProgramInfoLogFunc = NULL; //ericw
+QS_PFNGLCREATEPROGRAMPROC GL_CreateProgramFunc = NULL; //ericw
+QS_PFNGLATTACHSHADERPROC GL_AttachShaderFunc = NULL; //ericw
+QS_PFNGLLINKPROGRAMPROC GL_LinkProgramFunc = NULL; //ericw
+QS_PFNGLBINDATTRIBLOCATIONFUNC GL_BindAttribLocationFunc = NULL; //ericw
+QS_PFNGLUSEPROGRAMPROC GL_UseProgramFunc = NULL; //ericw
+QS_PFNGLGETATTRIBLOCATIONPROC GL_GetAttribLocationFunc = NULL; //ericw
+QS_PFNGLVERTEXATTRIBPOINTERPROC GL_VertexAttribPointerFunc = NULL; //ericw
+QS_PFNGLENABLEVERTEXATTRIBARRAYPROC GL_EnableVertexAttribArrayFunc = NULL; //ericw
+QS_PFNGLDISABLEVERTEXATTRIBARRAYPROC GL_DisableVertexAttribArrayFunc = NULL; //ericw
+QS_PFNGLGETUNIFORMLOCATIONPROC GL_GetUniformLocationFunc = NULL; //ericw
+QS_PFNGLUNIFORM1IPROC GL_Uniform1iFunc = NULL; //ericw
+QS_PFNGLUNIFORM1FPROC GL_Uniform1fFunc = NULL; //ericw
+QS_PFNGLUNIFORM3FPROC GL_Uniform3fFunc = NULL; //ericw
+QS_PFNGLUNIFORM4FPROC GL_Uniform4fFunc = NULL; //ericw
 
 //====================================
 
@@ -94,6 +147,8 @@ static cvar_t	vid_width = {"vid_width", "800", CVAR_ARCHIVE};		// QuakeSpasm, wa
 static cvar_t	vid_height = {"vid_height", "600", CVAR_ARCHIVE};	// QuakeSpasm, was 480
 static cvar_t	vid_bpp = {"vid_bpp", "16", CVAR_ARCHIVE};
 static cvar_t	vid_vsync = {"vid_vsync", "0", CVAR_ARCHIVE};
+static cvar_t	vid_fsaa = {"vid_fsaa", "0", CVAR_ARCHIVE}; // QuakeSpasm
+static cvar_t	vid_desktopfullscreen = {"vid_desktopfullscreen", "0", CVAR_ARCHIVE}; // QuakeSpasm
 //johnfitz
 
 cvar_t		vid_gamma = {"gamma", "1", CVAR_ARCHIVE}; //johnfitz -- moved here from view.c
@@ -104,6 +159,9 @@ cvar_t		vid_gamma = {"gamma", "1", CVAR_ARCHIVE}; //johnfitz -- moved here from
 //
 //==========================================================================
 
+#define	USE_GAMMA_RAMPS			0
+
+#if USE_GAMMA_RAMPS
 static unsigned short vid_gamma_red[256];
 static unsigned short vid_gamma_green[256];
 static unsigned short vid_gamma_blue[256];
@@ -111,8 +169,10 @@ static unsigned short vid_gamma_blue[256];
 static unsigned short vid_sysgamma_red[256];
 static unsigned short vid_sysgamma_green[256];
 static unsigned short vid_sysgamma_blue[256];
+#endif
 
-static int vid_gammaworks;
+static qboolean	gammaworks = false;	// whether hw-gamma works
+static int fsaa;
 
 /*
 ================
@@ -121,10 +181,35 @@ VID_Gamma_SetGamma -- apply gamma correction
 */
 static void VID_Gamma_SetGamma (void)
 {
-	if (draw_context && vid_gammaworks)
+	if (gl_glsl_gamma_able)
+		return;
+
+	if (draw_context && gammaworks)
 	{
-		if (SDL_SetGammaRamp(&vid_gamma_red[0], &vid_gamma_green[0], &vid_gamma_blue[0]) == -1)
+		float	value;
+
+		if (vid_gamma.value > (1.0f / GAMMA_MAX))
+			value = 1.0f / vid_gamma.value;
+		else
+			value = GAMMA_MAX;
+
+#if defined(USE_SDL2)
+# if USE_GAMMA_RAMPS
+		if (SDL_SetWindowGammaRamp(draw_context, vid_gamma_red, vid_gamma_green, vid_gamma_blue) != 0)
+			Con_Printf ("VID_Gamma_SetGamma: failed on SDL_SetWindowGammaRamp\n");
+# else
+		if (SDL_SetWindowBrightness(draw_context, value) != 0)
+			Con_Printf ("VID_Gamma_SetGamma: failed on SDL_SetWindowBrightness\n");
+# endif
+#else /* USE_SDL2 */
+# if USE_GAMMA_RAMPS
+		if (SDL_SetGammaRamp(vid_gamma_red, vid_gamma_green, vid_gamma_blue) == -1)
 			Con_Printf ("VID_Gamma_SetGamma: failed on SDL_SetGammaRamp\n");
+# else
+		if (SDL_SetGamma(value,value,value) == -1)
+			Con_Printf ("VID_Gamma_SetGamma: failed on SDL_SetGamma\n");
+# endif
+#endif /* USE_SDL2 */
 	}
 }
 
@@ -135,10 +220,28 @@ VID_Gamma_Restore -- restore system gamma
 */
 static void VID_Gamma_Restore (void)
 {
-	if (draw_context && vid_gammaworks)
+	if (gl_glsl_gamma_able)
+		return;
+
+	if (draw_context && gammaworks)
 	{
-		if (SDL_SetGammaRamp(&vid_sysgamma_red[0], &vid_sysgamma_green[0], &vid_sysgamma_blue[0]) == -1)
+#if defined(USE_SDL2)
+# if USE_GAMMA_RAMPS
+		if (SDL_SetWindowGammaRamp(draw_context, vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) != 0)
+			Con_Printf ("VID_Gamma_Restore: failed on SDL_SetWindowGammaRamp\n");
+# else
+		if (SDL_SetWindowBrightness(draw_context, 1) != 0)
+			Con_Printf ("VID_Gamma_Restore: failed on SDL_SetWindowBrightness\n");
+# endif
+#else /* USE_SDL2 */
+# if USE_GAMMA_RAMPS
+		if (SDL_SetGammaRamp(vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) == -1)
 			Con_Printf ("VID_Gamma_Restore: failed on SDL_SetGammaRamp\n");
+# else
+		if (SDL_SetGamma(1, 1, 1) == -1)
+			Con_Printf ("VID_Gamma_Restore: failed on SDL_SetGamma\n");
+# endif
+#endif /* USE_SDL2 */
 	}
 }
 
@@ -159,6 +262,10 @@ VID_Gamma_f -- callback when the cvar changes
 */
 static void VID_Gamma_f (cvar_t *var)
 {
+	if (gl_glsl_gamma_able)
+		return;
+
+#if USE_GAMMA_RAMPS
 	int i;
 
 	for (i = 0; i < 256; i++)
@@ -168,7 +275,7 @@ static void VID_Gamma_f (cvar_t *var)
 		vid_gamma_green[i] = vid_gamma_red[i];
 		vid_gamma_blue[i] = vid_gamma_red[i];
 	}
-
+#endif
 	VID_Gamma_SetGamma ();
 }
 
@@ -179,15 +286,203 @@ VID_Gamma_Init -- call on init
 */
 static void VID_Gamma_Init (void)
 {
-	vid_gammaworks = false;
-
-	if (SDL_GetGammaRamp (&vid_sysgamma_red[0], &vid_sysgamma_green[0], &vid_sysgamma_blue[0]) != -1)
-		vid_gammaworks = true;
-
 	Cvar_RegisterVariable (&vid_gamma);
 	Cvar_SetCallback (&vid_gamma, VID_Gamma_f);
+
+	if (gl_glsl_gamma_able)
+		return;
+
+#if defined(USE_SDL2)
+# if USE_GAMMA_RAMPS
+	gammaworks	= (SDL_GetWindowGammaRamp(draw_context, vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) == 0);
+	if (gammaworks)
+	    gammaworks	= (SDL_SetWindowGammaRamp(draw_context, vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) == 0);
+# else
+	gammaworks	= (SDL_SetWindowBrightness(draw_context, 1) == 0);
+# endif
+#else /* USE_SDL2 */
+# if USE_GAMMA_RAMPS
+	gammaworks	= (SDL_GetGammaRamp(vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) == 0);
+	if (gammaworks)
+	    gammaworks	= (SDL_SetGammaRamp(vid_sysgamma_red, vid_sysgamma_green, vid_sysgamma_blue) == 0);
+# else
+	gammaworks	= (SDL_SetGamma(1, 1, 1) == 0);
+# endif
+#endif /* USE_SDL2 */
+
+	if (!gammaworks)
+		Con_SafePrintf("gamma adjustment not available\n");
 }
 
+/*
+======================
+VID_GetCurrentWidth
+======================
+*/
+static int VID_GetCurrentWidth (void)
+{
+#if defined(USE_SDL2)
+	int w = 0, h = 0;
+	SDL_GetWindowSize(draw_context, &w, &h);
+	return w;
+#else
+	return draw_context->w;
+#endif
+}
+
+/*
+=======================
+VID_GetCurrentHeight
+=======================
+*/
+static int VID_GetCurrentHeight (void)
+{
+#if defined(USE_SDL2)
+	int w = 0, h = 0;
+	SDL_GetWindowSize(draw_context, &w, &h);
+	return h;
+#else
+	return draw_context->h;
+#endif
+}
+
+/*
+====================
+VID_GetCurrentBPP
+====================
+*/
+static int VID_GetCurrentBPP (void)
+{
+#if defined(USE_SDL2)
+	const Uint32 pixelFormat = SDL_GetWindowPixelFormat(draw_context);
+	return SDL_BITSPERPIXEL(pixelFormat);
+#else
+	return draw_context->format->BitsPerPixel;
+#endif
+}
+
+/*
+====================
+VID_GetFullscreen
+====================
+*/
+static qboolean VID_GetFullscreen (void)
+{
+#if defined(USE_SDL2)
+	return (SDL_GetWindowFlags(draw_context) & SDL_WINDOW_FULLSCREEN) != 0;
+#else
+	return (draw_context->flags & SDL_FULLSCREEN) != 0;
+#endif
+}
+
+/*
+====================
+VID_GetDesktopFullscreen
+====================
+*/
+static qboolean VID_GetDesktopFullscreen (void)
+{
+#if defined(USE_SDL2)
+	return (SDL_GetWindowFlags(draw_context) & SDL_WINDOW_FULLSCREEN_DESKTOP) != 0;
+#else
+	return false;
+#endif
+}
+
+/*
+====================
+VID_GetVSync
+====================
+*/
+static qboolean VID_GetVSync (void)
+{
+#if defined(USE_SDL2)
+	return SDL_GL_GetSwapInterval() == 1;
+#else
+	int swap_control;
+	if (SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &swap_control) == 0)
+		return swap_control > 0;
+	return false;
+#endif
+}
+
+/*
+====================
+VID_GetWindow
+
+used by pl_win.c
+====================
+*/
+void *VID_GetWindow (void)
+{
+#if defined(USE_SDL2)
+	return draw_context;
+#else
+	return NULL;
+#endif
+}
+
+/*
+====================
+VID_HasMouseOrInputFocus
+====================
+*/
+qboolean VID_HasMouseOrInputFocus (void)
+{
+#if defined(USE_SDL2)
+	return (SDL_GetWindowFlags(draw_context) & (SDL_WINDOW_MOUSE_FOCUS | SDL_WINDOW_INPUT_FOCUS)) != 0;
+#else
+	return (SDL_GetAppState() & (SDL_APPMOUSEFOCUS | SDL_APPINPUTFOCUS)) != 0;
+#endif
+}
+
+/*
+====================
+VID_IsMinimized
+====================
+*/
+qboolean VID_IsMinimized (void)
+{
+#if defined(USE_SDL2)
+	return !(SDL_GetWindowFlags(draw_context) & SDL_WINDOW_SHOWN);
+#else
+	/* SDL_APPACTIVE in SDL 1.x means "not minimized" */
+	return !(SDL_GetAppState() & SDL_APPACTIVE);
+#endif
+}
+
+#if defined(USE_SDL2)
+/*
+================
+VID_SDL2_GetDisplayMode
+
+Returns a pointer to a statically allocated SDL_DisplayMode structure
+if there is one with the requested params on the default display.
+Otherwise returns NULL.
+
+This is passed to SDL_SetWindowDisplayMode to specify a pixel format
+with the requested bpp. If we didn't care about bpp we could just pass NULL.
+================
+*/
+static SDL_DisplayMode *VID_SDL2_GetDisplayMode(int width, int height, int bpp)
+{
+	static SDL_DisplayMode mode;
+	const int sdlmodes = SDL_GetNumDisplayModes(0);
+	int i;
+
+	for (i = 0; i < sdlmodes; i++)
+	{
+		if (SDL_GetDisplayMode(0, i, &mode) == 0
+			&& mode.w == width && mode.h == height
+			&& SDL_BITSPERPIXEL(mode.format) == bpp)
+		{
+			return &mode;
+		}
+	}
+	return NULL;
+}
+#endif /* USE_SDL2 */
+
 /*
 ================
 VID_ValidMode
@@ -195,18 +490,28 @@ VID_ValidMode
 */
 static qboolean VID_ValidMode (int width, int height, int bpp, qboolean fullscreen)
 {
-	Uint32 flags = DEFAULT_SDL_FLAGS;
-
+// ignore width / height / bpp if vid_desktopfullscreen is enabled
+    if (fullscreen && vid_desktopfullscreen.value)
+        return true;
+    
 	if (width < 320)
 		return false;
 
 	if (height < 200)
 		return false;
 
-	if (fullscreen)
-		flags |= SDL_FULLSCREEN;
+#if defined(USE_SDL2)
+	if (fullscreen && VID_SDL2_GetDisplayMode(width, height, bpp) == NULL)
+		bpp = 0;
+#else
+	{
+		Uint32 flags = DEFAULT_SDL_FLAGS;
+		if (fullscreen)
+			flags |= SDL_FULLSCREEN;
 
-	bpp = SDL_VideoModeOK(width, height, bpp, flags);
+		bpp = SDL_VideoModeOK(width, height, bpp, flags);
+	}
+#endif
 
 	switch (bpp)
 	{
@@ -226,15 +531,14 @@ static qboolean VID_ValidMode (int width, int height, int bpp, qboolean fullscre
 VID_SetMode
 ================
 */
-static int VID_SetMode (int width, int height, int bpp, qboolean fullscreen)
+static qboolean VID_SetMode (int width, int height, int bpp, qboolean fullscreen)
 {
 	int		temp;
-	Uint32	flags = DEFAULT_SDL_FLAGS;
+	Uint32	flags;
 	char		caption[50];
-
-	if (fullscreen)
-		flags |= SDL_FULLSCREEN;
-
+	int		depthbits, stencilbits;
+	int		fsaa_obtained;
+	
 	// so Con_Printfs don't mess us up by forcing vid and snd updates
 	temp = scr_disabled_for_loading;
 	scr_disabled_for_loading = true;
@@ -242,9 +546,88 @@ static int VID_SetMode (int width, int height, int bpp, qboolean fullscreen)
 	CDAudio_Pause ();
 	BGM_Pause ();
 
-	//
-	// swap control (the "before SDL_SetVideoMode" part)
-	//
+	/* z-buffer depth */
+	if (bpp == 16)
+	{
+		depthbits = 16;
+		stencilbits = 0;
+	}
+	else
+	{
+		depthbits = 24;
+		stencilbits = 8;
+	}
+	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, depthbits);
+	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, stencilbits);
+
+	/* fsaa */
+	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, fsaa > 0 ? 1 : 0);
+	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, fsaa);
+
+	q_snprintf(caption, sizeof(caption), "QuakeSpasm %1.2f.%d", (float)QUAKESPASM_VERSION, QUAKESPASM_VER_PATCH);
+
+#if defined(USE_SDL2)
+	/* Create the window if needed, hidden */
+	if (!draw_context)
+	{
+		flags = SDL_WINDOW_OPENGL | SDL_WINDOW_HIDDEN;
+
+		draw_context = SDL_CreateWindow (caption, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, flags);
+		if (!draw_context) { // scale back fsaa
+			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 0);
+			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 0);
+			draw_context = SDL_CreateWindow (caption, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, flags);
+		}
+		if (!draw_context) { // scale back SDL_GL_DEPTH_SIZE
+			SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
+			draw_context = SDL_CreateWindow (caption, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, flags);
+		}
+		if (!draw_context) { // scale back SDL_GL_STENCIL_SIZE
+			SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 0);
+			draw_context = SDL_CreateWindow (caption, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, flags);
+		}
+		if (!draw_context)
+			Sys_Error ("Couldn't create window");
+
+		gl_context = SDL_GL_CreateContext (draw_context);
+		if (!gl_context)
+			Sys_Error ("Couldn't create GL context");
+	}
+
+	/* Ensure the window is not fullscreen */
+	if (VID_GetFullscreen ())
+	{
+		if (SDL_SetWindowFullscreen (draw_context, 0) != 0)
+			Sys_Error("Couldn't set fullscreen state mode");
+	}
+
+	/* Set window size and display mode */
+	SDL_SetWindowSize (draw_context, width, height);
+	SDL_SetWindowPosition (draw_context, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
+	SDL_SetWindowDisplayMode (draw_context, VID_SDL2_GetDisplayMode(width, height, bpp));
+
+	/* Make window fullscreen if needed, and show the window */
+
+	if (fullscreen) {
+		Uint32 flags = vid_desktopfullscreen.value ?
+			SDL_WINDOW_FULLSCREEN_DESKTOP :
+			SDL_WINDOW_FULLSCREEN;
+		if (SDL_SetWindowFullscreen (draw_context, flags) != 0)
+			Sys_Error ("Couldn't set fullscreen state mode");
+	}
+
+	SDL_ShowWindow (draw_context);
+
+	gl_swap_control = true;
+	if (SDL_GL_SetSwapInterval ((vid_vsync.value) ? 1 : 0) == -1)
+		gl_swap_control = false;
+
+#else /* !defined(USE_SDL2) */
+
+	flags = DEFAULT_SDL_FLAGS;
+	if (fullscreen)
+		flags |= SDL_FULLSCREEN;
+
 	gl_swap_control = true;
 	if (SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, (vid_vsync.value) ? 1 : 0) == -1)
 		gl_swap_control = false;
@@ -252,19 +635,44 @@ static int VID_SetMode (int width, int height, int bpp, qboolean fullscreen)
 	bpp = SDL_VideoModeOK(width, height, bpp, flags);
 
 	draw_context = SDL_SetVideoMode(width, height, bpp, flags);
-	if (!draw_context)
-		Sys_Error ("Couldn't set video mode");
+	if (!draw_context) { // scale back fsaa
+		SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 0);
+		SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 0);
+		draw_context = SDL_SetVideoMode(width, height, bpp, flags);
+	}
+	if (!draw_context) { // scale back SDL_GL_DEPTH_SIZE
+		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
+		draw_context = SDL_SetVideoMode(width, height, bpp, flags);
+	}
+	if (!draw_context) { // scale back SDL_GL_STENCIL_SIZE
+		SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 0);
+		draw_context = SDL_SetVideoMode(width, height, bpp, flags);
+		if (!draw_context)
+			Sys_Error ("Couldn't set video mode");
+	}
 
-	q_snprintf(caption, sizeof(caption), "QuakeSpasm %1.2f.%d", (float)FITZQUAKE_VERSION, QUAKESPASM_VER_PATCH);
 	SDL_WM_SetCaption(caption, caption);
+#endif /* !defined(USE_SDL2) */
 
-	vid.width = draw_context->w;
-	vid.height = draw_context->h;
+	vid.width = VID_GetCurrentWidth();
+	vid.height = VID_GetCurrentHeight();
 	vid.conwidth = vid.width & 0xFFFFFFF8;
 	vid.conheight = vid.conwidth * vid.height / vid.width;
 	vid.numpages = 2;
 
-	modestate = draw_context->flags & SDL_FULLSCREEN ? MS_FULLSCREEN : MS_WINDOWED;
+// read the obtained z-buffer depth
+	if (SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &depthbits) == -1)
+		depthbits = 0;
+
+// read obtained fsaa samples
+	if (SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES, &fsaa_obtained) == -1)
+		fsaa_obtained = 0;
+
+// read stencil bits
+	if (SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE, &gl_stencilbits) == -1)
+		gl_stencilbits = 0;
+
+	modestate = VID_GetFullscreen() ? MS_FULLSCREEN : MS_WINDOWED;
 
 	CDAudio_Resume ();
 	BGM_Resume ();
@@ -273,10 +681,12 @@ static int VID_SetMode (int width, int height, int bpp, qboolean fullscreen)
 // fix the leftover Alt from any Alt-Tab or the like that switched us away
 	ClearAllStates ();
 
-	Con_SafePrintf ("Video mode %dx%dx%d initialized\n",
-				draw_context->w,
-				draw_context->h,
-				draw_context->format->BitsPerPixel);
+	Con_SafePrintf ("Video mode %dx%dx%d (%d-bit z-buffer, %dx FSAA) initialized\n",
+				VID_GetCurrentWidth(),
+				VID_GetCurrentHeight(),
+				VID_GetCurrentBPP(),
+				depthbits,
+				fsaa_obtained);
 
 	vid.recalc_refdef = 1;
 
@@ -323,6 +733,15 @@ static void VID_Restart (void)
 				width, height, bpp, fullscreen? "fullscreen" : "windowed");
 		return;
 	}
+	
+// ericw -- depending on platform / SDL version, after a video mode change we
+// can have a new context (all textures are already freed) or the same context
+// as before, in which case we need to delete the old textures to avoid a
+// memory leak.
+
+	TexMgr_DeleteTextureObjects ();
+	GLSLGamma_DeleteTexture ();
+	R_DeleteShaders ();
 
 //
 // set new mode
@@ -331,6 +750,8 @@ static void VID_Restart (void)
 
 	GL_Init ();
 	TexMgr_ReloadImages ();
+	GL_BuildVBOs ();
+	GLMesh_LoadVertexBuffers ();
 	GL_SetupState ();
 
 	//warpimages needs to be recalculated
@@ -371,10 +792,10 @@ static void VID_Test (void)
 //
 // now try the switch
 //
-	old_width = draw_context->w;
-	old_height = draw_context->h;
-	old_bpp = draw_context->format->BitsPerPixel;
-	old_fullscreen = draw_context->flags & SDL_FULLSCREEN ? true : false;
+	old_width = VID_GetCurrentWidth();
+	old_height = VID_GetCurrentHeight();
+	old_bpp = VID_GetCurrentBPP();
+	old_fullscreen = VID_GetFullscreen() ? true : false;
 
 	VID_Restart ();
 
@@ -486,7 +907,30 @@ static void GL_CheckExtensions (void)
 {
 	int swap_control;
 
+	// ARB_vertex_buffer_object
 	//
+	if (COM_CheckParm("-novbo"))
+		Con_Warning ("Vertex buffer objects disabled at command line\n");
+	else if (gl_version_major < 1 || (gl_version_major == 1 && gl_version_minor < 5))
+		Con_Warning ("OpenGL version < 1.5, skipping ARB_vertex_buffer_object check\n");
+	else
+	{
+		GL_BindBufferFunc = (PFNGLBINDBUFFERARBPROC) SDL_GL_GetProcAddress("glBindBufferARB");
+		GL_BufferDataFunc = (PFNGLBUFFERDATAARBPROC) SDL_GL_GetProcAddress("glBufferDataARB");
+		GL_BufferSubDataFunc = (PFNGLBUFFERSUBDATAARBPROC) SDL_GL_GetProcAddress("glBufferSubDataARB");
+		GL_DeleteBuffersFunc = (PFNGLDELETEBUFFERSARBPROC) SDL_GL_GetProcAddress("glDeleteBuffersARB");
+		GL_GenBuffersFunc = (PFNGLGENBUFFERSARBPROC) SDL_GL_GetProcAddress("glGenBuffersARB");
+		if (GL_BindBufferFunc && GL_BufferDataFunc && GL_BufferSubDataFunc && GL_DeleteBuffersFunc && GL_GenBuffersFunc)
+		{
+			Con_Printf("FOUND: ARB_vertex_buffer_object\n");
+			gl_vbo_able = true;
+		}
+		else
+		{
+			Con_Warning ("ARB_vertex_buffer_object not available\n");
+		}
+	}
+
 	// multitexture
 	//
 	if (COM_CheckParm("-nomtex"))
@@ -495,28 +939,14 @@ static void GL_CheckExtensions (void)
 	{
 		GL_MTexCoord2fFunc = (PFNGLMULTITEXCOORD2FARBPROC) SDL_GL_GetProcAddress("glMultiTexCoord2fARB");
 		GL_SelectTextureFunc = (PFNGLACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress("glActiveTextureARB");
-		if (GL_MTexCoord2fFunc && GL_SelectTextureFunc)
+		GL_ClientActiveTextureFunc = (PFNGLCLIENTACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress("glClientActiveTextureARB");
+		if (GL_MTexCoord2fFunc && GL_SelectTextureFunc && GL_ClientActiveTextureFunc)
 		{
 			Con_Printf("FOUND: ARB_multitexture\n");
-			TEXTURE0 = GL_TEXTURE0_ARB;
-			TEXTURE1 = GL_TEXTURE1_ARB;
-			gl_mtexable = true;
-		}
-		else
-		{
-			Con_Warning ("Couldn't link to multitexture functions\n");
-		}
-	}
-	else if (GL_ParseExtensionList(gl_extensions, "GL_SGIS_multitexture"))
-	{
-		GL_MTexCoord2fFunc = (PFNGLMULTITEXCOORD2FARBPROC) SDL_GL_GetProcAddress("glMTexCoord2fSGIS");
-		GL_SelectTextureFunc = (PFNGLACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress("glSelectTextureSGIS");
-		if (GL_MTexCoord2fFunc && GL_SelectTextureFunc)
-		{
-			Con_Printf("FOUND: SGIS_multitexture\n");
-			TEXTURE0 = TEXTURE0_SGIS;
-			TEXTURE1 = TEXTURE1_SGIS;
 			gl_mtexable = true;
+			
+			glGetIntegerv(GL_MAX_TEXTURE_UNITS, &gl_max_texture_units);
+			Con_Printf("GL_MAX_TEXTURE_UNITS: %d\n", (int)gl_max_texture_units);
 		}
 		else
 		{
@@ -528,7 +958,6 @@ static void GL_CheckExtensions (void)
 		Con_Warning ("multitexture not supported (extension not found)\n");
 	}
 
-	//
 	// texture_env_combine
 	//
 	if (COM_CheckParm("-nocombine"))
@@ -548,7 +977,6 @@ static void GL_CheckExtensions (void)
 		Con_Warning ("texture_env_combine not supported\n");
 	}
 
-	//
 	// texture_env_add
 	//
 	if (COM_CheckParm("-noadd"))
@@ -568,17 +996,28 @@ static void GL_CheckExtensions (void)
 		Con_Warning ("texture_env_add not supported\n");
 	}
 
-	//
-	// swap control (the "after SDL_SetVideoMode" part)
+	// swap control
 	//
 	if (!gl_swap_control)
 	{
+#if defined(USE_SDL2)
+		Con_Warning ("vertical sync not supported (SDL_GL_SetSwapInterval failed)\n");
+#else
 		Con_Warning ("vertical sync not supported (SDL_GL_SetAttribute failed)\n");
+#endif
 	}
+#if defined(USE_SDL2)
+	else if ((swap_control = SDL_GL_GetSwapInterval()) == -1)
+#else
 	else if (SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &swap_control) == -1)
+#endif
 	{
 		gl_swap_control = false;
+#if defined(USE_SDL2)
+		Con_Warning ("vertical sync not supported (SDL_GL_GetSwapInterval failed)\n");
+#else
 		Con_Warning ("vertical sync not supported (SDL_GL_GetAttribute failed)\n");
+#endif
 	}
 	else if ((vid_vsync.value && swap_control != 1) || (!vid_vsync.value && swap_control != 0))
 	{
@@ -587,10 +1026,13 @@ static void GL_CheckExtensions (void)
 	}
 	else
 	{
+#if defined(USE_SDL2)
+		Con_Printf("FOUND: SDL_GL_SetSwapInterval\n");
+#else
 		Con_Printf("FOUND: SDL_GL_SWAP_CONTROL\n");
+#endif
 	}
 
-	//
 	// anisotropic filtering
 	//
 	if (GL_ParseExtensionList(gl_extensions, "GL_EXT_texture_filter_anisotropic"))
@@ -632,6 +1074,113 @@ static void GL_CheckExtensions (void)
 		gl_max_anisotropy = 1;
 		Con_Warning ("texture_filter_anisotropic not supported\n");
 	}
+
+	// texture_non_power_of_two
+	//
+	if (COM_CheckParm("-notexturenpot"))
+		Con_Warning ("texture_non_power_of_two disabled at command line\n");
+	else if (GL_ParseExtensionList(gl_extensions, "GL_ARB_texture_non_power_of_two"))
+	{
+		Con_Printf("FOUND: ARB_texture_non_power_of_two\n");
+		gl_texture_NPOT = true;
+	}
+	else
+	{
+		Con_Warning ("texture_non_power_of_two not supported\n");
+	}
+	
+	// GLSL
+	//
+	if (COM_CheckParm("-noglsl"))
+		Con_Warning ("GLSL disabled at command line\n");
+	else if (gl_version_major >= 2)
+	{
+		GL_CreateShaderFunc = (QS_PFNGLCREATESHADERPROC) SDL_GL_GetProcAddress("glCreateShader");
+		GL_DeleteShaderFunc = (QS_PFNGLDELETESHADERPROC) SDL_GL_GetProcAddress("glDeleteShader");
+		GL_DeleteProgramFunc = (QS_PFNGLDELETEPROGRAMPROC) SDL_GL_GetProcAddress("glDeleteProgram");
+		GL_ShaderSourceFunc = (QS_PFNGLSHADERSOURCEPROC) SDL_GL_GetProcAddress("glShaderSource");
+		GL_CompileShaderFunc = (QS_PFNGLCOMPILESHADERPROC) SDL_GL_GetProcAddress("glCompileShader");
+		GL_GetShaderivFunc = (QS_PFNGLGETSHADERIVPROC) SDL_GL_GetProcAddress("glGetShaderiv");
+		GL_GetShaderInfoLogFunc = (QS_PFNGLGETSHADERINFOLOGPROC) SDL_GL_GetProcAddress("glGetShaderInfoLog");
+		GL_GetProgramivFunc = (QS_PFNGLGETPROGRAMIVPROC) SDL_GL_GetProcAddress("glGetProgramiv");
+		GL_GetProgramInfoLogFunc = (QS_PFNGLGETPROGRAMINFOLOGPROC) SDL_GL_GetProcAddress("glGetProgramInfoLog");
+		GL_CreateProgramFunc = (QS_PFNGLCREATEPROGRAMPROC) SDL_GL_GetProcAddress("glCreateProgram");
+		GL_AttachShaderFunc = (QS_PFNGLATTACHSHADERPROC) SDL_GL_GetProcAddress("glAttachShader");
+		GL_LinkProgramFunc = (QS_PFNGLLINKPROGRAMPROC) SDL_GL_GetProcAddress("glLinkProgram");
+		GL_BindAttribLocationFunc = (QS_PFNGLBINDATTRIBLOCATIONFUNC) SDL_GL_GetProcAddress("glBindAttribLocation");
+		GL_UseProgramFunc = (QS_PFNGLUSEPROGRAMPROC) SDL_GL_GetProcAddress("glUseProgram");
+		GL_GetAttribLocationFunc = (QS_PFNGLGETATTRIBLOCATIONPROC) SDL_GL_GetProcAddress("glGetAttribLocation");
+		GL_VertexAttribPointerFunc = (QS_PFNGLVERTEXATTRIBPOINTERPROC) SDL_GL_GetProcAddress("glVertexAttribPointer");
+		GL_EnableVertexAttribArrayFunc = (QS_PFNGLENABLEVERTEXATTRIBARRAYPROC) SDL_GL_GetProcAddress("glEnableVertexAttribArray");
+		GL_DisableVertexAttribArrayFunc = (QS_PFNGLDISABLEVERTEXATTRIBARRAYPROC) SDL_GL_GetProcAddress("glDisableVertexAttribArray");
+		GL_GetUniformLocationFunc = (QS_PFNGLGETUNIFORMLOCATIONPROC) SDL_GL_GetProcAddress("glGetUniformLocation");
+		GL_Uniform1iFunc = (QS_PFNGLUNIFORM1IPROC) SDL_GL_GetProcAddress("glUniform1i");
+		GL_Uniform1fFunc = (QS_PFNGLUNIFORM1FPROC) SDL_GL_GetProcAddress("glUniform1f");
+		GL_Uniform3fFunc = (QS_PFNGLUNIFORM3FPROC) SDL_GL_GetProcAddress("glUniform3f");
+		GL_Uniform4fFunc = (QS_PFNGLUNIFORM4FPROC) SDL_GL_GetProcAddress("glUniform4f");
+
+		if (GL_CreateShaderFunc &&
+			GL_DeleteShaderFunc &&
+			GL_DeleteProgramFunc &&
+			GL_ShaderSourceFunc &&
+			GL_CompileShaderFunc &&
+			GL_GetShaderivFunc &&
+			GL_GetShaderInfoLogFunc &&
+			GL_GetProgramivFunc &&
+			GL_GetProgramInfoLogFunc &&
+			GL_CreateProgramFunc &&
+			GL_AttachShaderFunc &&
+			GL_LinkProgramFunc &&
+			GL_BindAttribLocationFunc &&
+			GL_UseProgramFunc &&
+			GL_GetAttribLocationFunc &&
+			GL_VertexAttribPointerFunc &&
+			GL_EnableVertexAttribArrayFunc &&
+			GL_DisableVertexAttribArrayFunc &&
+			GL_GetUniformLocationFunc &&
+			GL_Uniform1iFunc &&
+			GL_Uniform1fFunc &&
+			GL_Uniform3fFunc &&
+			GL_Uniform4fFunc)
+		{
+			Con_Printf("FOUND: GLSL\n");
+			gl_glsl_able = true;
+		}
+		else
+		{
+			Con_Warning ("GLSL not available\n");
+		}
+	}
+	else
+	{
+		Con_Warning ("OpenGL version < 2, GLSL not available\n");
+	}
+	
+	// GLSL gamma
+	//
+	if (COM_CheckParm("-noglslgamma"))
+		Con_Warning ("GLSL gamma disabled at command line\n");
+	else if (gl_glsl_able && gl_texture_NPOT)
+	{
+		gl_glsl_gamma_able = true;
+	}
+	else
+	{
+		Con_Warning ("GLSL gamma not available, using hardware gamma\n");
+	}
+    
+    // GLSL alias model rendering
+    //
+	if (COM_CheckParm("-noglslalias"))
+		Con_Warning ("GLSL alias model rendering disabled at command line\n");
+	else if (gl_glsl_able && gl_vbo_able && gl_max_texture_units >= 3)
+	{
+		gl_glsl_alias_able = true;
+	}
+	else
+	{
+		Con_Warning ("GLSL alias model rendering not available, using Fitz renderer\n");
+	}
 }
 
 /*
@@ -674,12 +1223,28 @@ static void GL_Init (void)
 	gl_version = (const char *) glGetString (GL_VERSION);
 	gl_extensions = (const char *) glGetString (GL_EXTENSIONS);
 
+	if (gl_version == NULL || sscanf(gl_version, "%d.%d", &gl_version_major, &gl_version_minor) < 2)
+	{
+		gl_version_major = 0;
+		gl_version_minor = 0;
+	}
+
 	if (gl_extensions_nice != NULL)
 		Z_Free (gl_extensions_nice);
 	gl_extensions_nice = GL_MakeNiceExtensionsList (gl_extensions);
 
 	GL_CheckExtensions (); //johnfitz
 
+#ifdef __APPLE__
+	// ericw -- enable multi-threaded OpenGL, gives a decent FPS boost.
+	// https://developer.apple.com/library/mac/technotes/tn2085/
+	if (host_parms->numcpus > 1 &&
+	    kCGLNoError != CGLEnable(CGLGetCurrentContext(), kCGLCEMPEngine))
+	{
+		Con_Warning ("Couldn't enable multi-threaded OpenGL");
+	}
+#endif
+
 	//johnfitz -- intel video workarounds from Baker
 	if (!strcmp(gl_vendor, "Intel"))
 	{
@@ -687,6 +1252,9 @@ static void GL_Init (void)
 		Cbuf_AddText ("gl_clear 1");
 	}
 	//johnfitz
+
+	GLAlias_CreateShaders ();
+	GL_ClearBufferBindings ();	
 }
 
 /*
@@ -709,7 +1277,13 @@ GL_EndRendering
 void GL_EndRendering (void)
 {
 	if (!scr_skipupdate)
+	{
+#if defined(USE_SDL2)
+		SDL_GL_SwapWindow(draw_context);
+#else
 		SDL_GL_SwapBuffers();
+#endif
+	}
 }
 
 
@@ -721,13 +1295,13 @@ void	VID_Shutdown (void)
 
 		SDL_QuitSubSystem(SDL_INIT_VIDEO);
 		draw_context = NULL;
-
+#if defined(USE_SDL2)
+		gl_context = NULL;
+#endif
 		PL_VID_Shutdown();
 	}
 }
 
-//==========================================================================
-
 /*
 ===================================================================
 
@@ -748,7 +1322,6 @@ static void ClearAllStates (void)
 }
 
 
-
 //==========================================================================
 //
 //  COMMANDS
@@ -764,10 +1337,10 @@ static void VID_DescribeCurrentMode_f (void)
 {
 	if (draw_context)
 		Con_Printf("%dx%dx%d %s\n",
-			draw_context->w,
-			draw_context->h,
-			draw_context->format->BitsPerPixel,
-			draw_context->flags & SDL_FULLSCREEN ? "fullscreen" : "windowed");
+			VID_GetCurrentWidth(),
+			VID_GetCurrentHeight(),
+			VID_GetCurrentBPP(),
+			VID_GetFullscreen() ? "fullscreen" : "windowed");
 }
 
 /*
@@ -798,6 +1371,18 @@ static void VID_DescribeModes_f (void)
 	Con_Printf ("\n%i modes\n", count);
 }
 
+/*
+===================
+VID_FSAA_f -- ericw -- warn that vid_fsaa requires engine restart
+===================
+*/
+static void VID_FSAA_f (cvar_t *var)
+{
+	// don't print the warning if vid_fsaa is set during startup
+	if (vid_initialized)
+		Con_Printf("%s %d requires engine restart to take effect\n", var->name, (int)var->value);
+}
+
 //==========================================================================
 //
 //  INIT
@@ -811,6 +1396,26 @@ VID_InitModelist
 */
 static void VID_InitModelist (void)
 {
+#if defined(USE_SDL2)
+	const int sdlmodes = SDL_GetNumDisplayModes(0);
+	int i;
+
+	nummodes = 0;
+	for (i = 0; i < sdlmodes; i++)
+	{
+		SDL_DisplayMode mode;
+
+		if (nummodes >= MAX_MODE_LIST)
+			break;
+		if (SDL_GetDisplayMode(0, i, &mode) == 0)
+		{
+			modelist[nummodes].width = mode.w;
+			modelist[nummodes].height = mode.h;
+			modelist[nummodes].bpp = SDL_BITSPERPIXEL(mode.format);
+			nummodes++;
+		}
+	}
+#else /* !defined(USE_SDL2) */
 	SDL_PixelFormat	format;
 	SDL_Rect	**modes;
 	Uint32		flags;
@@ -862,6 +1467,7 @@ static void VID_InitModelist (void)
 
 	if (nummodes == originalnummodes)
 		Con_SafePrintf ("No fullscreen DIB modes found\n");
+#endif /* !defined(USE_SDL2) */
 }
 
 /*
@@ -872,14 +1478,15 @@ VID_Init
 void	VID_Init (void)
 {
 	static char vid_center[] = "SDL_VIDEO_CENTERED=center";
-	const SDL_VideoInfo *info;
-	int		width, height, bpp;
+	int		p, width, height, bpp, display_width, display_height, display_bpp;
 	qboolean	fullscreen;
 	const char	*read_vars[] = { "vid_fullscreen",
 					 "vid_width",
 					 "vid_height",
 					 "vid_bpp",
-					 "vid_vsync" };
+					 "vid_vsync",
+					 "vid_fsaa",
+					 "vid_desktopfullscreen" };
 #define num_readvars	( sizeof(read_vars)/sizeof(read_vars[0]) )
 
 	Cvar_RegisterVariable (&vid_fullscreen); //johnfitz
@@ -887,11 +1494,15 @@ void	VID_Init (void)
 	Cvar_RegisterVariable (&vid_height); //johnfitz
 	Cvar_RegisterVariable (&vid_bpp); //johnfitz
 	Cvar_RegisterVariable (&vid_vsync); //johnfitz
+	Cvar_RegisterVariable (&vid_fsaa); //QuakeSpasm
+	Cvar_RegisterVariable (&vid_desktopfullscreen); //QuakeSpasm
 	Cvar_SetCallback (&vid_fullscreen, VID_Changed_f);
 	Cvar_SetCallback (&vid_width, VID_Changed_f);
 	Cvar_SetCallback (&vid_height, VID_Changed_f);
 	Cvar_SetCallback (&vid_bpp, VID_Changed_f);
 	Cvar_SetCallback (&vid_vsync, VID_Changed_f);
+	Cvar_SetCallback (&vid_fsaa, VID_FSAA_f);
+	Cvar_SetCallback (&vid_desktopfullscreen, VID_Changed_f);
 
 	Cmd_AddCommand ("vid_unlock", VID_Unlock); //johnfitz
 	Cmd_AddCommand ("vid_restart", VID_Restart); //johnfitz
@@ -901,11 +1512,29 @@ void	VID_Init (void)
 
 	putenv (vid_center);	/* SDL_putenv is problematic in versions <= 1.2.9 */
 
-	if (SDL_InitSubSystem(SDL_INIT_VIDEO) == -1)
-		Sys_Error("Could not initialize SDL Video");
+	if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0)
+		Sys_Error("Couldn't init SDL video: %s", SDL_GetError());
 
-	info = SDL_GetVideoInfo();
-	Cvar_SetValueQuick (&vid_bpp, (float)info->vfmt->BitsPerPixel);
+#if defined(USE_SDL2)
+	{
+		SDL_DisplayMode mode;
+		if (SDL_GetDesktopDisplayMode(0, &mode) != 0)
+			Sys_Error("Could not get desktop display mode");
+
+		display_width = mode.w;
+		display_height = mode.h;
+		display_bpp = SDL_BITSPERPIXEL(mode.format);
+	}
+#else
+	{
+		const SDL_VideoInfo *info = SDL_GetVideoInfo();
+		display_width = info->current_w;
+		display_height = info->current_h;
+		display_bpp = info->vfmt->BitsPerPixel;
+	}
+#endif
+
+	Cvar_SetValueQuick (&vid_bpp, (float)display_bpp);
 
 	if (CFG_OpenConfig("config.cfg") == 0)
 	{
@@ -920,18 +1549,17 @@ void	VID_Init (void)
 	height = (int)vid_height.value;
 	bpp = (int)vid_bpp.value;
 	fullscreen = (int)vid_fullscreen.value;
+	fsaa = (int)vid_fsaa.value;
 
 	if (COM_CheckParm("-current"))
 	{
-		width = info->current_w;
-		height = info->current_h;
-		bpp = info->vfmt->BitsPerPixel;
+		width = display_width;
+		height = display_height;
+		bpp = display_bpp;
 		fullscreen = true;
 	}
 	else
 	{
-		int p;
-
 		p = COM_CheckParm("-width");
 		if (p && p < com_argc-1)
 		{
@@ -960,6 +1588,10 @@ void	VID_Init (void)
 			fullscreen = true;
 	}
 
+	p = COM_CheckParm ("-fsaa");
+	if (p && p < com_argc-1)
+		fsaa = atoi(com_argv[p+1]);
+
 	if (!VID_ValidMode(width, height, bpp, fullscreen))
 	{
 		width = (int)vid_width.value;
@@ -972,7 +1604,7 @@ void	VID_Init (void)
 	{
 		width = 640;
 		height = 480;
-		bpp = info->vfmt->BitsPerPixel;
+		bpp = display_bpp;
 		fullscreen = false;
 	}
 
@@ -1010,16 +1642,41 @@ void	VID_Init (void)
 void	VID_Toggle (void)
 {
 	static qboolean vid_toggle_works = true;
+	qboolean toggleWorked;
+	Uint32 flags = 0;
 
 	S_ClearBuffer ();
 
 	if (!vid_toggle_works)
 		goto vrestart;
-	if (SDL_WM_ToggleFullScreen(draw_context) == 1)
+	else if (gl_vbo_able)
+	{
+		// disabling the fast path because with SDL 1.2 it invalidates VBOs (using them
+		// causes a crash, sugesting that the fullscreen toggle created a new GL context,
+		// although texture objects remain valid for some reason).
+		//
+		// SDL2 does promise window resizes / fullscreen changes preserve the GL context,
+		// so we could use the fast path with SDL2. --ericw
+		vid_toggle_works = false;
+		goto vrestart;
+	}
+
+#if defined(USE_SDL2)
+	if (!VID_GetFullscreen())
+	{
+		flags = vid_desktopfullscreen.value ? SDL_WINDOW_FULLSCREEN_DESKTOP : SDL_WINDOW_FULLSCREEN;
+	}
+
+	toggleWorked = SDL_SetWindowFullscreen(draw_context, flags) == 0;
+#else
+	toggleWorked = SDL_WM_ToggleFullScreen(draw_context) == 1;
+#endif
+
+	if (toggleWorked)
 	{
 		Sbar_Changed ();	// Sbar seems to need refreshing
 
-		modestate = draw_context->flags & SDL_FULLSCREEN ? MS_FULLSCREEN : MS_WINDOWED;
+		modestate = VID_GetFullscreen() ? MS_FULLSCREEN : MS_WINDOWED;
 
 		VID_SyncCvars();
 
@@ -1037,7 +1694,7 @@ void	VID_Toggle (void)
 		vid_toggle_works = false;
 		Con_DPrintf ("SDL_WM_ToggleFullScreen failed, attempting VID_Restart\n");
 	vrestart:
-		Cvar_SetQuick (&vid_fullscreen, draw_context-> flags & SDL_FULLSCREEN ? "0" : "1");
+		Cvar_SetQuick (&vid_fullscreen, VID_GetFullscreen() ? "0" : "1");
 		Cbuf_AddText ("vid_restart\n");
 	}
 }
@@ -1049,17 +1706,16 @@ VID_SyncCvars -- johnfitz -- set vid cvars to match current video mode
 */
 void VID_SyncCvars (void)
 {
-	int swap_control;
-
 	if (draw_context)
 	{
-		Cvar_SetValueQuick (&vid_width, draw_context->w);
-		Cvar_SetValueQuick (&vid_height, draw_context->h);
-		Cvar_SetValueQuick (&vid_bpp, draw_context->format->BitsPerPixel);
-		Cvar_SetQuick (&vid_fullscreen, draw_context->flags & SDL_FULLSCREEN ? "1" : "0");
-
-		if (SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &swap_control) == 0)
-			Cvar_SetQuick (&vid_vsync, (swap_control > 0)? "1" : "0");
+        if (!VID_GetDesktopFullscreen())
+        {
+            Cvar_SetValueQuick (&vid_width, VID_GetCurrentWidth());
+            Cvar_SetValueQuick (&vid_height, VID_GetCurrentHeight());
+        }
+		Cvar_SetValueQuick (&vid_bpp, VID_GetCurrentBPP());
+		Cvar_SetQuick (&vid_fullscreen, VID_GetFullscreen() ? "1" : "0");
+		Cvar_SetQuick (&vid_vsync, VID_GetVSync() ? "1" : "0");
 	}
 
 	vid_changed = false;
@@ -1326,6 +1982,7 @@ static void VID_MenuKey (int key)
 		break;
 
 	case K_ENTER:
+	case K_KP_ENTER:
 		m_entersound = true;
 		switch (video_options_cursor)
 		{
