@@ -2,6 +2,7 @@
 Copyright (C) 1996-2001 Id Software, Inc.
 Copyright (C) 2002-2009 John Fitzgibbons and others
 Copyright (C) 2007-2008 Kristian Duske
+Copyright (C) 2010-2014 QuakeSpasm developers
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -219,7 +220,7 @@ void	Host_FindMaxClients (void)
 void Host_Version_f (void)
 {
 	Con_Printf ("Quake Version %1.2f\n", VERSION);
-	Con_Printf ("QuakeSpasm Version %1.2f.%d\n", FITZQUAKE_VERSION, QUAKESPASM_VER_PATCH);
+	Con_Printf ("QuakeSpasm Version %1.2f.%d\n", QUAKESPASM_VERSION, QUAKESPASM_VER_PATCH);
 	Con_Printf ("Exe: "__TIME__" "__DATE__"\n");
 }
 
@@ -653,7 +654,7 @@ void Host_ServerFrame (void)
 				active++;
 		}
 		if (active > 600 && dev_peakstats.edicts <= 600)
-			Con_Warning ("%i edicts exceeds standard limit of 600.\n", active);
+			Con_DWarning ("%i edicts exceeds standard limit of 600.\n", active);
 		dev_stats.edicts = active;
 		dev_peakstats.edicts = q_max(active, dev_peakstats.edicts);
 	}
@@ -689,6 +690,7 @@ void _Host_Frame (float time)
 
 // get new key events
 	Key_UpdateForDest ();
+	IN_UpdateInputMode ();
 	Sys_SendKeyEvents ();
 
 // allow mice or other external controllers to add commands
@@ -856,6 +858,7 @@ void Host_Init (void)
 		M_Init ();
 		ExtraMaps_Init (); //johnfitz
 		Modlist_Init (); //johnfitz
+		DemoList_Init (); //ericw
 		VID_Init ();
 		IN_Init ();
 		TexMgr_Init (); //johnfitz
